package jq.qde.constraint;

import jq.qde.QDEState;
import jq.qde.QualitativeConstraint;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.value.QualitativeDirection;
import jq.qde.space.value.QualitativeValue;
import jq.qsim.QSimListener;

/**
 * x+y=z
 * @author oursland
 */
public class MultConstraint extends QualitativeConstraint {

	public MultConstraint(QualitativeVariable[] args, ConstraintCorrespondence[] corr) {
		super("mult", args, corr);
	}

	public MultConstraint(QualitativeVariable x, QualitativeVariable y, QualitativeVariable z) {
		super("mult", new QualitativeVariable[] { x, y, z }, EMPTYCORR);
	}

	private static final int INC = QualitativeDirection.INC_MASK;
	private static final int STD = QualitativeDirection.STD_MASK;
	private static final int DEC = QualitativeDirection.DEC_MASK;
	private static final int ANY = QualitativeDirection.ANY_MASK;

	/*  multrel[rx][ry] contains the allowable relations of z to 0 in z = x * y,
	*  where rx and ry are th relations of x and y to 0, repsectively.
	*  This array is ordered on the assumption that GT=0, EQ=1, LT=2.
	*/
	private static final int[][] multrel = { 
		{ INC, STD, DEC }, 
		{ STD, STD, STD }, 
		{ DEC, STD, INC } 
	};
	public static int getRelationMask(QualitativeDirection s1, QualitativeDirection s2) {
		return multrel[s1.key][s2.key];
	}


	/*
	 *  The following arrays contain the legal directions-of-change for the
	 *  various combinations of signs of f, g and h in MULT(f,g,h). 
	 *  The arrays are ordered on the assumption that INC=0, EQ=1, DEC=2.
	 *  The array names follow a pattern -- they all begin "fgh", and the
	 *  next 3 characters indicate the relations of f, g and h to 0,
	 *  where 'p' = +, 'z' = 0, and 'm' = -.
	 */
	private static final int[][] fghppp = 	{{INC, INC, ANY}, {INC, STD, DEC}, {ANY, DEC, DEC}};
	private static final int[][] fghpmm =	{{ANY, DEC, DEC}, {INC, STD, DEC}, {INC, INC, ANY}};
	private static final int[][] fghpzz =	{{INC, STD, DEC}, {INC, STD, DEC}, {INC, STD, DEC}};
	private static final int[][] fghzpz =	{{INC, INC, INC}, {STD, STD, STD}, {DEC, DEC, DEC}};
	private static final int[][] fghzzz =	{{STD, STD, STD}, {STD, STD, STD}, {STD, STD, STD}};
	private static final int[][] fghzmz =	{{DEC, DEC, DEC}, {STD, STD, STD}, {INC, INC, INC}};
	private static final int[][] fghmpm =	{{ANY, INC, INC}, {DEC, STD, INC}, {DEC, DEC, ANY}};
	private static final int[][] fghmzz =	{{DEC, STD, INC}, {DEC, STD, INC}, {DEC, STD, INC}};
	private static final int[][] fghmmp =	{{DEC, DEC, ANY}, {DEC, STD, INC}, {ANY, INC, INC}};
	
	/*
	 *  multdir[xsign][ysign][zsign] contains a pointer to the appropriate table
	 *  of combinations-of-directions-of-change.  xsign, ysign, and zsign are
	 *  the relations of x, y and z to 0, respectively.  This array is ordered
	 *  on the assumption that GT=0, EQ=1, LT=2.
	 */
	private static final int[][][][][] multdir = 
	{	{ { fghppp, null, null }, { null, fghpzz, null }, { null, null, fghpmm } },
		{ { null, fghzpz, null }, { null, fghzzz, null }, { null, fghzmz, null } },
		{ { null, null, fghmpm }, { null, fghmzz, null }, { fghmmp, null, null } }
	};

	public ConstraintValidationResult validateValues(QualitativeValue[] stateArray) {
		final QualitativeDirection s0 = stateArray[0].getMagnitudeSign();
		final QualitativeDirection s1 = stateArray[1].getMagnitudeSign();
		final QualitativeDirection s2 = stateArray[2].getMagnitudeSign();

		/*  Check for legal combinations of relations to zero  */
		if ((s2.mask & getRelationMask(s0, s1)) == 0) {
			return new ConstraintValidationResult(false, "zero relation", this, stateArray);
		}

		/*  Qualitative directions must agree.  */
		final int[][] multp = multdir[s0.key][s1.key][s2.key];
		final QualitativeDirection d0 = stateArray[0].direction;
		final QualitativeDirection d1 = stateArray[1].direction;
		final QualitativeDirection d2 = stateArray[2].direction;
		if( (multp[d0.key][d1.key] & d2.mask) == 0 ) {
			return new ConstraintValidationResult(false, "qdir mismatch", this, stateArray);
		}

		/*  Loop over all corresponding values for this constraint */
		for( int i = 0; i < getCorrespondingValueCount(); i++ ) {
			final ConstraintCorrespondence corr = getCorrespondingValue(i);
			final QualitativeDirection c0 = stateArray[0].compareMagnitude(corr.getValue(stateArray[0]));
			final QualitativeDirection c1 = stateArray[1].compareMagnitude(corr.getValue(stateArray[1]));
			final QualitativeDirection c2 = stateArray[2].compareMagnitude(corr.getValue(stateArray[2]));
			if ((multp[c0.key][c1.key] & c2.mask) == 0) {
				return new ConstraintValidationResult(false, "cval mismatch", this, stateArray);
			}
		}

		/*  Check infinite values  */
		final QualitativeDirection i0 = stateArray[0].getMagnitudeInfiniteSign();
		final QualitativeDirection i1 = stateArray[1].getMagnitudeInfiniteSign();
		final QualitativeDirection i2 = stateArray[2].getMagnitudeInfiniteSign();
		if((i2.mask & AddConstraint.getQDirMask(i0, i1)) == 0) {
			return new ConstraintValidationResult(false, "fails infinite check", this, stateArray);
		}
		return new ConstraintValidationResult(true, "OK", this, stateArray);
	}

	public boolean shouldCreateNewCorrespondence(QDEState state, QSimListener l) {
		return !state.getVariableMagnitude(getVariable(2)).isZero() && super.shouldCreateNewCorrespondence(state, l);
	}

	public QualitativeConstraint createCopy(QualitativeVariable[] args, ConstraintCorrespondence[] corr) {
		return new MultConstraint(args, corr);
	}
}
