package jq.qde.space;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;

import jq.qde.space.value.MagnitudeInterval;
import jq.qde.space.value.MagnitudeLandmark;
import jq.qde.space.value.QualitativeDirection;
import jq.qde.space.value.QualitativeMagnitude;
import jq.qde.space.value.QualitativeValue;
import jq.util.Abbrevable;

/**
 * @author oursland
 */
public class VariableQuantitySpace implements QuantitySpace, Abbrevable {
	public final QualitativeVariable variable;
	private final MagnitudeLandmark[] landmarkArray;
	private final MagnitudeInterval[] intervalArray;
	private final HashMap nodeMap = new HashMap();
	private final QSpaceNode low; 
	private final QSpaceNode high; 
	private final HashMap signMap = new HashMap();

	public VariableQuantitySpace(QualitativeVariable var, String[] lms) {
		this(var, var.getLandmarks(lms));
	}
	
	public VariableQuantitySpace(QualitativeVariable var, MagnitudeLandmark[] lms) {
		this.variable = var;
		this.landmarkArray = (MagnitudeLandmark[]) lms.clone();
		this.intervalArray = new MagnitudeInterval[lms.length-1];
		for (int i = 0; i < lms.length-1; i++) {
			this.intervalArray[i] = var.getInterval(lms[i], lms[i+1]);
		}
		if( lms.length > 0 ) {
			assert lms[0].getVariable() == var;
			QSpaceNode prev = createNode(lms[0], null);
			this.low = prev;
			for (int i = 1; i < lms.length; i++) {
				assert lms[i].getVariable() == var;
				QSpaceNode intervalNode = createNode(this.intervalArray[i-1], prev);
				prev.setNext(intervalNode);

				QSpaceNode curr = createNode(lms[i], intervalNode);
				intervalNode.setNext(curr);

				prev = curr;
			}
			this.high = prev;
		} else {
			this.low = null;
			this.high = null;
		}
	}

	private QSpaceNode createNode(QualitativeMagnitude mag, QSpaceNode prevNode) {
		QSpaceNode currNode = new QSpaceNode(prevNode, mag);
		nodeMap.put(mag, currNode);
		if( prevNode != null ) {
			prevNode.setNext(currNode);
		}
		return currNode;
	}

	public QuantitySpace addAfter(MagnitudeLandmark prevLM, MagnitudeLandmark newLM) {
		MagnitudeLandmark[] newLandmarkArray = getNewLandmarkArray(prevLM, newLM);
		return new VariableQuantitySpace(variable, newLandmarkArray);
	}

	protected MagnitudeLandmark[] getNewLandmarkArray(MagnitudeLandmark prevLM, MagnitudeLandmark newLM) {
		assert !nodeMap.containsKey(newLM);
		assert nodeMap.containsKey(prevLM);
		MagnitudeLandmark[] newLandmarkArray = new MagnitudeLandmark[landmarkArray.length+1];
		copyLandmarks(prevLM, newLM, newLandmarkArray);
//		ValueInterval[] newIntervalArray = new ValueInterval[newLandmarkArray.length-1];
//		for (int i = 0; i < newIntervalArray.length; i++) {
//			newIntervalArray[i] = var.getInterval(newLandmarkArray[i], newLandmarkArray[i+1]);
//		}
		return newLandmarkArray;
	}

	private void copyLandmarks(MagnitudeLandmark prevLM, MagnitudeLandmark newLM, MagnitudeLandmark[] newLandmarkArray) {
		int src = 0;
		int dst = 0;
		if( prevLM != null ) {
			do {
				newLandmarkArray[dst++] = landmarkArray[src++];
			} while( landmarkArray[src-1] != prevLM );			
		}
		newLandmarkArray[dst++] = newLM;
		while( src < landmarkArray.length ) {
			newLandmarkArray[dst++] = landmarkArray[src++];
		}
	}

	public Iterator landmarkIterator() {
		return Arrays.asList(landmarkArray).iterator();
	}

	public ListIterator landmarkListIterator() {
		return Arrays.asList(landmarkArray).listIterator();
	}

	private MagnitudeLandmark getLandmark(int i) {
		return landmarkArray[i];
	}

	private MagnitudeInterval getInterval(int i) {
		return intervalArray[i];
	}

	public MagnitudeLandmark getLandmarkValue(String name) {
		for (int i = 0; i < landmarkArray.length; i++) {
			if( getLandmark(i).getName().equals(name) ) {
				return getLandmark(i);
			}
		}
		throw new Error("Landmark " + name + " does not exist in variable " + variable.getName());
	}
	
	public boolean lessThan(MagnitudeLandmark value, MagnitudeLandmark high) {
		if( value == high ) {
			return false;
		}
		Iterator iter = landmarkListIterator(); 
		while(iter.hasNext()) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			if( lm == value ) {
				break;
			}
			if( lm == high ) {
				return false;
			}
		}
		while(iter.hasNext()) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			if( lm == high ) {
				return true;
			}
		}		
		return false;
	}
	
	public int getQValueCount() {
		return nodeMap.size();
	}
	
	public QualitativeMagnitude getQValue( int index ) {
		int lowIndex = index/2;
		if( index%2 == 0 ) {
			// landmark value
			return getLandmark(lowIndex);
		} else {
			// interval value
			return getInterval(lowIndex);
		}
	}

	public void printSpecification(PrintStream out, String prefix) {
		out.print(prefix);
		out.print( "(" + variable.getName() + " (");
		for (Iterator iter = landmarkIterator(); iter.hasNext();) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			out.print(lm.getName());			
			if( iter.hasNext() ) {
				out.print(" ");
			}
		}
		out.print("))");
	}

	public QualitativeDirection compare(QualitativeMagnitude magnitude, QualitativeMagnitude landmark) {
		if( magnitude == landmark ) {
			return QualitativeDirection.ZERO;
		} else if( lessThan(magnitude, landmark) ) {
			return QualitativeDirection.NEGATIVE;
		} else if( greaterThan(magnitude, landmark) ) {
			return QualitativeDirection.POSITIVE;
		}
		return QualitativeDirection.UNKNOWN;
	}

	private boolean lessThan(QualitativeMagnitude low, QualitativeMagnitude high) {
		final boolean containsLow = nodeMap.containsKey(low);
		final boolean containsHigh = nodeMap.containsKey(high);
		final int qLow = getQSpaceIndex(low);
		final int qHigh = getQSpaceIndex(high);
		if( containsLow && containsHigh && qLow < qHigh ) {
			return true;
		} else {
			return false;
		}
	}

	private boolean greaterThan(QualitativeMagnitude low, QualitativeMagnitude high) {
		final boolean containsLow = nodeMap.containsKey(low);
		final boolean containsHigh = nodeMap.containsKey(high);
		final int qLow = getQSpaceIndex(low);
		final int qHigh = getQSpaceIndex(high);
		if( containsLow && containsHigh && qLow > qHigh ) {
			return true;
		} else {
			return false;
		}
	}

	public QualitativeDirection getSign(QualitativeMagnitude magnitude) {
		if( magnitude.isZero() ) {
			return QualitativeDirection.ZERO;
		} else if( magnitude.isInfinite() ) {
			return magnitude.getInfiniteSign();
		} else {
			QualitativeDirection sign = (QualitativeDirection) signMap.get(magnitude);
			if( sign == null ) {
				sign = compare(magnitude, magnitude.getVariable().ZERO);
				signMap.put(magnitude, sign);
			}
			return sign;
		}
	}

	public boolean isNegative(QualitativeMagnitude mag) {
		return getSign(mag).isDecreasing();
	}

	public boolean isPositive(QualitativeMagnitude mag) {
		return getSign(mag).isIncreasing();
	}

	public boolean isLowEnd(QualitativeMagnitude magnitude) {
		return magnitude == low.mag;
	}

	public boolean isHighEnd(QualitativeMagnitude magnitude) {
		return magnitude == high.mag;
	}

	private QualitativeMagnitude getNext(QualitativeMagnitude mag) {
		QSpaceNode node = (QSpaceNode) nodeMap.get(mag);
		return node.next.mag;
	}

	private QualitativeMagnitude getPrev(QualitativeMagnitude mag) {
		QSpaceNode node = (QSpaceNode) nodeMap.get(mag);
		return node.prev.mag;
	}

	public QualitativeMagnitude shiftMagnitude(QualitativeMagnitude mag, int i) {
		while( i > 0 ) {
			mag = getNext(mag);
			i--;
		}
		while( i < 0 ) {
			mag = getPrev(mag);
			i++;
		}
		return mag;
	}

	public int getQSpaceIndex(QualitativeMagnitude mag) {
		assert mag.getVariable() == variable;
		if( !mag.isUnknown() ) {
			QSpaceNode node = (QSpaceNode) nodeMap.get(mag);
			return node.qspaceIndex;
		} else {
			throw new Error("Unknown values do not have qspace indexes.");
		}
	}

	public boolean isDiverging(QualitativeValue v) {
		final boolean divergingDown =  isLowEnd(v.magnitude) && v.direction.isDecreasing();
		final boolean divergingUp =   isHighEnd(v.magnitude) && v.direction.isIncreasing();
		return divergingDown || divergingUp;
	}

	private static class QSpaceNode {
		final int qspaceIndex;
		final QSpaceNode prev;
		QSpaceNode next;
		final QualitativeMagnitude mag;
		
		public QSpaceNode(QSpaceNode prev, QualitativeMagnitude landmark) {
			if( prev != null ) {
				qspaceIndex = prev.qspaceIndex+1;
			} else {
				qspaceIndex = 0;
			}
			this.prev = prev;
			this.next = null;
			this.mag = landmark;
		}
		public void setNext(QSpaceNode next) {
			assert qspaceIndex == next.qspaceIndex-1;
			this.next = next;			
		}
	}
	public boolean isEnd(QualitativeMagnitude mag) {
		return isHighEnd(mag) && isLowEnd(mag);
	}

	public String toSpecification() {
		StringBuffer rc = new StringBuffer();
		rc.append("(");
		rc.append(variable.getName());
		rc.append("\t");
		rc.append(toAbbreviation());
		rc.append(")");
		return rc.toString();
	}

	public String toAbbreviation() {
		StringBuffer rc = new StringBuffer();
		rc.append("(");
		for (int i = 0; i < landmarkArray.length; i++) {
			if( i > 0 ) {
				rc.append(" ");
			}
			rc.append(landmarkArray[i].toAbbreviation());
		}
		rc.append(")");
		return rc.toString();
	}

	public int getLandmarkCount() {
		return landmarkArray.length;
	}

	public boolean contains(MagnitudeLandmark landmark) {
		for (int i = 0; i < landmarkArray.length; i++) {
			if( landmarkArray[i] == landmark ) {
				return true;
			}
		}
		return false;
	}
}
