package jq.qde.space.value;

import jq.qde.space.QualitativeVariable;
import jq.qde.space.VariableQuantitySpace;
import jq.qsim.QSimListener;
import jq.qsim.VariablePossibleValues;
import jq.util.Abbrevable;


/**
 * @author oursland
 */
public class QualitativeValue implements Abbrevable {
	public final VariableQuantitySpace qspace;
	public final QualitativeVariable variable;
	public final QualitativeMagnitude magnitude;
	public final QualitativeDirection direction;

	public QualitativeValue(VariableQuantitySpace qspace) {
		this.qspace = qspace;
		this.variable = qspace.variable;
		this.magnitude = qspace.variable.UNKNOWN;
		this.direction = QualitativeDirection.UNKNOWN;
	}

	public QualitativeValue(VariableQuantitySpace qspace, QualitativeMagnitude value, QualitativeDirection direction) {
		this.qspace = qspace;
		this.variable = qspace.variable;
		this.magnitude = value;
		this.direction = direction;
	}

	public String toSpecification() {
		return "(" + variable.getName() + " (" + magnitude.toSpecification() + " " + direction.toDirectionString() + "))";
	}

	public String toAbbreviation() {
		return "(" + magnitude.toAbbreviation() + " " + direction.toDirectionString() + ")";
	}

	//	public void nextStates(Collection results) {
	//		final SignValue INC = SignValue.INC;
	//		final SignValue DEC = SignValue.DEC;
	//		final SignValue STD = SignValue.STD;
	//		if( direction.isSteady() ) {
	//			addState(results, variable, magnitude, STD);	// no qualitative change
	//			addState(results, variable, magnitude, INC);	// the value begins increasing
	//			addState(results, variable, magnitude, DEC);	// the value begins decreasing
	//		} else if( direction.isIncreasing() || direction.isDecreasing() ) {
	//			addState(results, variable, magnitude.nextInterval(direction), direction);	// no qualitative change
	//			addState(results, variable, magnitude.nextInterval(direction), STD);		// stops at a new landmark value
	//			addState(results, variable, magnitude.nextLandmark(direction), direction);	// reaches the next landmark value
	//			addState(results, variable, magnitude.nextLandmark(direction), STD);		// stops at the next landmark value
	//		} else if( direction.isUnknown() ) {
	//			new QualitativeValue(variable, magnitude, STD).nextStates(results);
	//			new QualitativeValue(variable, magnitude, INC).nextStates(results);
	//			new QualitativeValue(variable, magnitude, DEC).nextStates(results);
	//		}
	//	}

//	private static final void addState(
//		Collection results,
//		VariableQuantitySpace qspace,
//		QualitativeVariable variable,
//		QualitativeMagnitude value,
//		QualitativeDirection direction) {
//		if( variable == null)
//			return;
//		if( value == null)
//			return;
//		if( direction == null)
//			return;
//		QualitativeValue state = new QualitativeValue(qspace, variable, value, direction);
//		results.add(state);
//	}

	public VariablePossibleValues getInitialValues() {
		VariablePossibleValues rc = new VariablePossibleValues(qspace);
		QualitativeMagnitude[] possibleMags = magnitude.getInitialValues(qspace);
		QualitativeDirection[] possibleSigns = direction.getPossibleCurrentValues();
		for (int i = 0; i < possibleMags.length; i++) {
			for (int j = 0; j < possibleSigns.length; j++) {
				rc.add(new QualitativeValue(qspace, possibleMags[i], possibleSigns[j]));
			}
		}
		return rc;
	}

	public VariablePossibleValues getPossibleValues(TimeValue time, QSimListener l) {
		l.creatingPossibleValues(this);
		VariablePossibleValues rc = new VariablePossibleValues(qspace);

		final int tpi = time.isInterval() ? 1 : 0; // time & 0x01;
		final QualitativeVariable var = getVariable();

		if( var.isIndependent() ) {
			/*  This is an independent parameter, so use its initial value */
			rc.add(new QualitativeValue(qspace, magnitude, direction));
			return rc;
		}

		QValDelta[] transp;
		if( magnitude.isInterval() ) {
			/*  Magnitude is an interval, so test for adjacency to minf/inf  */
			MagnitudeInterval interval = (MagnitudeInterval) magnitude;
			if( interval.low.isInfinite() ) {
				/*  qmag is adjacent to minf (-infinity)  */
				transp = trans1[tpi][direction.key];
			} else if( interval.high.isInfinite() ) {
				/*  qmag is adjacent to inf (+infinity)  */
				transp = trans2[tpi][direction.key];
			} else {
				/*  qmag is somewhere in the middle of its qspace  */
				transp = trans3[tpi][direction.key];
			}
		} else {
			/*  Magnitude is a point, so test for upper/lower limits  */
			if( isMagnitudeEnd()) {
				/*  qmag is at lower limit of qspace  */
				transp = trans4[tpi][direction.key];
			} else if( isMagnitudeEnd() ) {
				/*  qmag is at upper limit of qspace  */
				transp = trans5[tpi][direction.key];
			} else {
				/*  qmag is somewhere in the middle of its qspace  */
				transp = trans6[tpi][direction.key];
			}
		}

		/*  'transp' now points to the set of legal transitions.  */
		if( transp.length == 0) {
			throw new Error("Invalid transition attempted on param " + var.getName());
		}

		/*  Loop over each transition and install the pvals  */
		for( int j = 0; j < transp.length; j++ ) {
			rc.add(new QualitativeValue(qspace, qspace.shiftMagnitude(magnitude, transp[j].qmagdiff), transp[j].qdir));
		}

		return rc;
	}

	private boolean isMagnitudeEnd() {
		return qspace.isEnd(magnitude);
	}

	public QualitativeVariable getVariable() {
		return variable;
	}
	
	private static class QValDelta {
		final int qmagdiff;
		final QualitativeDirection qdir;
		public QValDelta(int qmagdiff, QualitativeDirection qdir) {
			this.qmagdiff = qmagdiff;
			this.qdir = qdir;
		}
	}
	
	private static final QualitativeDirection INC = QualitativeDirection.INC;
	private static final QualitativeDirection STD = QualitativeDirection.STD;
	private static final QualitativeDirection DEC = QualitativeDirection.DEC;
	
	/*
	 *  The P-transitions and I-transitions that follow are as defined in
	 *  "Qualitative Simulation" by Benjamin Kuipers, Artificial Intelligence,
	 *  Vol. 29, page 300.  Each definition contains an increment for the 
	 *  qmag index and a new qdir value.
	 *
	 *				    QS(f,t)         ==>  QS(f,t,t+1)
	 *				__________________________________________
	 */
	private static final QValDelta P1 = new QValDelta( 0, STD );		/*  <l,std>         ==>  <l,std>	*/
	private static final QValDelta P2 = new QValDelta( 1, INC );		/*  <l,std>	    ==>  <(l,l+1),inc>	*/
	private static final QValDelta P3 = new QValDelta(-1, DEC );		/*  <l,std>	    ==>  <(l-1,l),dec>	*/
	private static final QValDelta P4 = new QValDelta( 1, INC );		/*  <l,inc>	    ==>  <(l,l+1),inc>	*/
	private static final QValDelta P5 = new QValDelta( 0, INC );		/*  <(l,l+1),inc>   ==>  <(l,l+1),inc>	*/
	private static final QValDelta P6 = new QValDelta(-1, DEC );		/*  <l,dec>	    ==>  <(l-1,l),dec>	*/
	private static final QValDelta P7 = new QValDelta( 0, DEC );		/*  <(l,l+1),dec>   ==>  <(l,l+1),dec>	*/

	/*
	 *				    QS(f,t,t+1)     ==>  QS(f,t+1)
	 *				__________________________________________
	 */
	private static final QValDelta I1 = new QValDelta( 0, STD );		/*  <l,std>	    ==>  <l,std>	*/
	private static final QValDelta I2 = new QValDelta( 1, STD );		/*  <(l,l+1),inc>   ==>  <l+1,std>	*/
	private static final QValDelta I3 = new QValDelta( 1, INC );		/*  <(l,l+1),inc>   ==>  <l+1,inc>	*/
	private static final QValDelta I4 = new QValDelta( 0, INC );		/*  <(l,l+1),inc>   ==>  <(l,l+1),inc>	*/
	private static final QValDelta I5 = new QValDelta(-1, STD );		/*  <(l,l+1),dec>   ==>  <l,std>	*/
	private static final QValDelta I6 = new QValDelta(-1, DEC );		/*  <(l,l+1),dec>   ==>  <l,dec>	*/
	private static final QValDelta I7 = new QValDelta( 0, DEC );		/*  <(l,l+1),dec>   ==>  <(l,l+1),dec>	*/
	private static final QValDelta I8 = new QValDelta( 0, STD );		/*  <(l,l+1),inc>   ==>  <l*,std>	*/
	private static final QValDelta I9 = new QValDelta( 0, STD );		/*  <(l,l+1),dec>   ==>  <l*,std>	*/

	private static final QValDelta trans1[][][] =
	{   { /* ================ Time is a point =============================== */
			{P5},	/*  <(l,l+1), inc> */
			{},			/*  <(l,l+1), std> note 1 */
			{P7},	/*  <(l,l+1), dec> */
		},
		{ /* ================ Time is an interval =========================== */
			{I2, I3, I4, I8},	/*  <(l,l+1), inc>	*/
			{},						/*  <(l,l+1), std> note 1 */
			{I7, I9}				/*  <(l,l+1), dec>	*/
		}
	};


	private static final QValDelta trans2[][][] =
	{   { /* ================ Time is a point =============================== */
			{P5},	/*  <(l,l+1), inc>	*/
			{},			/*  <(l,l+1), std> note 1 */
			{P7},	/*  <(l,l+1), dec>	*/
		},
		{ /* ================ Time is an interval =========================== */
			{I4, I8},		/*  <(l,l+1), inc>	*/
			{},					/*  <(l,l+1), std> note 1 */
			{I5, I6, I7, I9} /*  <(l,l+1), dec>	*/
		}
	};


	private static final QValDelta trans3[][][] =
	{   { /* ================ Time is a point =============================== */
			{P5},	/*  <(l,l+1), inc>	*/
			{},			/*  <(l,l+1), std> note 1 */
			{P7},	/*  <(l,l+1), dec>	*/
		},
		{ /* ================ Time is an interval =========================== */
			{I2, I3, I4, I8},	/*  <(l,l+1), inc>	*/
			{},						/*  <(l,l+1), std> note 1 */
			{I5, I6, I7, I9}		/*  <(l,l+1), dec>	*/
		}
	};


	private static final QValDelta trans4[][][] =
	{   { /* ================ Time is a point ============================= */
			{P4},		/*  <l, inc>		*/
			{P1, P2},	/*  <l, std>		*/
			{},				/*  <l, dec>		*/
		},
		{ /* ================ Time is an interval ========================= */
			{},			/*  <l, inc>		*/
			{I1},	/*  <l, std>		*/
			{}			/*  <l, dec>		*/
		}
	};


	private static final QValDelta trans5[][][] =
	{   { /* ================ Time is a point ============================= */
			{},				/*  <l, inc>		*/
			{P1, P3},	/*  <l, std>		*/
			{P6},		/*  <l, dec>		*/
		},
		{ /* ================ Time is an interval ========================= */
			{},			/*  <l, inc>		*/
			{I1},	/*  <l, std>		*/
			{}			/*  <l, dec>		*/
		}
	};


	private static final QValDelta trans6[][][] =
	{   { /* ================ Time is a point =============================== */
			{P4},			/*  <l, inc>		*/
			{P1, P2, P3},	/*  <l, std>		*/
			{P6},			/*  <l, dec>		*/
		},
		{ /* ================ Time is an interval =========================== */
			{},			/*  <l, inc>		*/
			{I1},	/*  <l, std>		*/
			{}			/*  <l, dec>		*/
		}
	};
	
	private boolean hashCodeSet = false;
	private int hashCode = 0;
	public int hashCode() {
		if( !hashCodeSet ) {
			if( !magnitude.isUnknown()) {
				hashCode = 16*direction.key + qspace.getQSpaceIndex(magnitude);
			} else {
				hashCode = 16*direction.key + qspace.getQValueCount();				
			}
			hashCodeSet = true;
		}
		return hashCode;
	}

	public QualitativeDirection getMagnitudeSign() {
		return qspace.getSign(magnitude);
	}

	public QualitativeDirection compareMagnitude(MagnitudeLandmark landmark) {
		return qspace.compare(magnitude, landmark);
	}

	public QualitativeDirection getMagnitudeInfiniteSign() {
		QualitativeMagnitude magnitude1 = magnitude;
		return magnitude1.getInfiniteSign();
	}

	/**
	 * Compares the magnitude and directions of the states. Requires that there is a single
	 * instance of each landmark and interval value.
	 * @param value The value to compare.
	 * @return 'true' if the magnitude and direction are identical. 'false' otherwise.
	 */
	public boolean isQualitativelyEqual(QualitativeValue value) {
		return magnitude == value.magnitude && direction == value.direction;
	}

	public boolean isNewLandmarkCandidate() {
		return magnitude.isInterval() && direction.isSteady();
	}

	public boolean isSteady() {
		return direction.isSteady();
	}
}
