package jq.qsim;

import java.io.PrintStream;
import java.util.Iterator;
import oursland.text.PrintStreamColumnFormatter;

import jq.qde.QDEState;
import jq.qde.QualitativeConstraint;
import jq.qde.constraint.ConstraintCorrespondence;
import jq.qde.constraint.ConstraintValidationResult;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.VariableQuantitySpace;
import jq.qde.space.value.QualitativeValue;
import jq.qde.space.value.TimeValue;
import jq.qsim.globalfilter.GlobalFilter;
import jq.util.PrintUtil;

/**
 * @author oursland
 */
public class QSimPrinter implements QSimListener {
	private final PrintStream out;
	public final PrintStreamColumnFormatter stateFormatter;
	public final PrintStreamColumnFormatter constraintFormatter;
	public final PrintStreamColumnFormatter corrFormatter;

	public QSimPrinter(PrintStream out) {
		this.out = out;
		this.stateFormatter = new PrintStreamColumnFormatter(out, new int[] {4, 16, 32});
		//0123456789012345678901234567890123456789
		//     Parameter   Value            Quantity Space

		this.constraintFormatter = new PrintStreamColumnFormatter(out, new int[] {0, 8, 24, 40, 56, 72});
		//d//dt   y               v
		//01234567890123456789012345
		
		this.corrFormatter = new PrintStreamColumnFormatter(out, new int[] {0, 8, 24, 40, 56, 72});
		//add     0               0               0
		//01234567890123456789012345678901234567890123456789

	}

	private void printConstraint(QualitativeConstraint constraint) {
		constraintFormatter.nextColumn(constraint.getName());
		for (int i = 0; i < constraint.getVariableCount(); i++) {
			constraintFormatter.nextColumn(constraint.getVariable(i).getName());
		}
		constraintFormatter.nextLine();
	}

	QualitativeConstraint lastConstraint = null;
	public void createTuple(ValueTuple tuple, QualitativeConstraint constraint) {
		if( lastConstraint != constraint ) {
			lastConstraint =  constraint;
			printConstraint(constraint);
		}
		constraintFormatter.nextColumn(constraint.getName());
		for (int i = 0; i < constraint.getVariableCount(); i++) {
			QualitativeValue value = tuple.getValue(i);
			constraintFormatter.nextColumn("(" + value.magnitude.toSpecification() + " " + value.direction.toDirectionString() + ")");
		}
		constraintFormatter.out.print("  OK");
		constraintFormatter.nextLine();
	}

	public void rejectTuple(QualitativeValue[] tuple, QualitativeConstraint constraint, ConstraintValidationResult validation) {
		if( lastConstraint != constraint ) {
			lastConstraint =  constraint;
			printConstraint(constraint);
		}
		constraintFormatter.nextColumn(constraint.getName());
		for (int i = 0; i < constraint.getVariableCount(); i++) {
			QualitativeValue value = tuple[i];
			constraintFormatter.nextColumn("(" + value.magnitude.toSpecification() + " " + value.direction.toDirectionString() + ")");
		}
		constraintFormatter.out.print("  " + validation.msg);
		constraintFormatter.nextLine();
	}


	public void filterTuple(ValueTuple tuple, QualitativeConstraint constraint, boolean accepted, QSimFilter filter) {
		if( lastConstraint != constraint ) {
			lastConstraint =  constraint;
			printConstraint(constraint);
		}
		constraintFormatter.nextColumn(constraint.getName());
		for (int i = 0; i < constraint.getVariableCount(); i++) {
			QualitativeValue value = tuple.getValue(i);
			constraintFormatter.nextColumn("(" + value.magnitude.toSpecification() + " " + value.direction .toDirectionString() + ")");
		}
		for (int i = 0; i < constraint.getVariableCount(); i++) {
			QualitativeValue value = tuple.getValue(i);
			out.print(" [" + filter.getValueUseCount(value) + " " + filter.getVariableUseCount(value.variable) + "]");
		}
		out.print((accepted? "" : " rejected"));
		constraintFormatter.nextLine();
		
	}

	public void checkConstraintArgument(
		QualitativeConstraint constraint,
		QualitativeValue[] value,
		ConstraintValidationResult result) {
		// TODO Auto-generated method stub - checkConstraintArgument

	}

	public void initialState(QDEState state) {
		// TODO Auto-generated method stub - initialState

	}

	public void startPrediction(QDEState state) {
		out.println("Predicting from state: " + state.getIdentifier());
	}

	public void initialValues(VariablePossibleValues ivals) {
		out.println(ivals.getVariable().getName() + ":  " + ivals.getPossibleValueCount() + " pvals:");
		Iterator iter = ivals.valueIterator();
		while(iter.hasNext()) {
			QualitativeValue value = (QualitativeValue) iter.next();
			out.print("        ");
			out.print(value.toAbbreviation());
			out.println();
		}
	}

	public void creatingPossibleValues(QualitativeValue value) {
		final QualitativeVariable var = value.variable;
		final VariableQuantitySpace qspace = value.qspace;
		final String name = var.getName();
		final int qSpaceIndex = qspace.getQSpaceIndex(value.magnitude);
		final int landmarkCount = qspace.getLandmarkCount();
		out.println("genpvals: parameter "+name+", qmagx="+qSpaceIndex+", # landmarks="+landmarkCount);
	}

	public void possibleValues(VariablePossibleValues pvals) {
		initialValues(pvals);
	}

	public void filterTupleOut(ValueTuple tuple, QualitativeConstraint constraint) {
		out.print("Tuple Filtered Out ");
		out.print(tuple.toSpecification());
		out.println();
	}

	public void allFilteredOut(QualitativeConstraint constraint) {
		out.print("All tuples Waltz-filtered-out for constraint ");
		constraint.printSpecification(out, "");
		out.println();
	}

	public void tentativeState(QDEState candidateState, GlobalFilter failedFilter) {
		// TODO: Why is the initial state at time (-1 0)? Tentative state 0, time = (-1 0), nochange = 0
		int id = candidateState.getIdentifier();
		TimeValue time = candidateState.getTime();
		out.println("Tentative state " + id + ", time = " + time.toAbbreviation() + ", filter = " + (failedFilter!=null?failedFilter.getName():"OK"));
//		PrintUtil.printState(stateFormatter, candidateState);
	}

	public void mismatchState(ValueTuple tuple, QDEState state) {
//		out.print("Tuple ");
//		out.print(tuple.toAbbreviation());
//		out.print(" contradicts: ");
//		out.print(state.toAbbreviation());
//		out.println();
	}

	public void formedState(QDEState state) {
		PrintUtil.printState(stateFormatter, state);
	}

	public void inheritingCorrespondence(QualitativeConstraint conp, ConstraintCorrespondence corrp) {
		corrFormatter.nextColumn(conp.getName());
		for( int i = 0; i < corrp.getCorrespondingValueCount(); i++ ) {
			corrFormatter.nextColumn(corrp.getValue(i).toAbbreviation());
		}
		corrFormatter.nextLine();
	}

	public void addingCorrespondence(QualitativeConstraint conp) {
		out.println("Before adding correspondence, ncorrs="+conp.getCorrespondingValueCount());
	}

	public void existingCorrespondence(QualitativeConstraint conp, ConstraintCorrespondence corr) {
//		out.println("Current qmagx's are exact match for:");
//		PrintUtil.printCorrespondence(out, corr);
	}

	public void addedCorrespondence(QualitativeConstraint conp, ConstraintCorrespondence corr) {
		out.println("New correspondence below, ncorrs="+conp.getCorrespondingValueCount()+", crp=?");
		PrintUtil.printCorrespondence(out, corr);
	}

	public void assignTupleValue(ValueTuple tuple, QualitativeValue value) {
//		out.println("Assigning tuple " + value.getVariable().getName() + "=" + value.toAbbreviation() + " from " + tuple.constraint.getName());
	}

	public void quiescentState(QDEState parentState) {
		out.println("State " + parentState.getIdentifier() + " is quiescent, so it has no children.");
	}

	public void divergingState(QDEState parentState) {
		out.println("State " + parentState.getIdentifier() + " is diverging, so it has no children.");
	}
}
