package jq.qsim.globalfilter;

import oursland.naming.NamableObject;
import jq.qde.QDEState;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.value.QualitativeMagnitude;
import jq.qde.space.value.QualitativeValue;

/**
 * @author oursland
 */
public class CycleFilter extends NamableObject implements GlobalFilter {

	public CycleFilter() {
		super("CycleFilter: state matches a predecessor");
	}

	/**
	 *  CHECK FOR CYCLE:
	 *  If the candidate state is at a time point and it has no new landmarks
	 *  and it has a predecessor (i.e., it is not the starting state), then
	 *  check through its predecessors (at time points only) for an identical
	 *  state.  If found, update predecessor and successor lists to show that
	 *  a cycle has occurred (and discard the candidate state).
	 */
	public boolean passesFilter(QDEState candidateState, QDEState previousState) {
		// no previous state mean no cycle
		// change in qspace means no cycle
		return (previousState == null) || 
		       (candidateState.qspace != previousState.qspace) || 
		       !hasCycle(candidateState, previousState);
	}

	private boolean hasCycle(QDEState candidateState, QDEState previousState) {
		boolean cycleFound = false;
		final QDEState prevp = getPreviousMatch(candidateState);
		if (prevp != null) {
			/*  Candidate state at this time point matches a predecessor
			 *  at an earlier time point.  So, indicate a cycle by making
			 *  previous state a successor of current state and by making
			 *  current state a predecessor of previous state.  Note that:
			 *  -- previousState points to current state,
			 *  -- statep points to the candidate state (which is
			 *     an immediate successor of current state), and
			 *  -- prevp points the matching previous state.
			 */
			// we will not for a new state in this case
			previousState.addSuccessor(prevp);
			prevp.addPredecessor(previousState);
			cycleFound = true;
		}
		return cycleFound;
	}

	private QDEState getPreviousMatch(QDEState candidateState) {
		if (candidateState.getTime().isPoint()) {
			int statehash = candidateState.hashCode();
			QDEState prev = candidateState;
			out : 
			while (prev.getPredecessorCount() > 0) {
				prev = prev.getPredecessor(0);
				if (statehash == prev.hashCode()) {
					/*  Loop over all variables, test for exact match */
					if (candidateState.qspace != prev.qspace) {
						return null; // do not need to check any more previous states
					} else if( isExactMatch( candidateState, prev) ) {
						return prev;
					}
				}
			}
		}
		return null;
	}

	private boolean isExactMatch(QDEState candidateState, QDEState prev) {
		final int variableCount = candidateState.qspace.getVariableCount();
		for (int i = 0; i < variableCount; i++) {
			final QualitativeVariable var = candidateState.qspace.getVariable(i);
			final QualitativeValue currqvalp = candidateState.getValue(var);
			final QualitativeValue prevqvalp = prev.getValue(var);
			final QualitativeMagnitude magnitude = currqvalp.magnitude;
			if (currqvalp.direction != prevqvalp.direction) {
				return false;
			}
			if (magnitude.isInterval()) {
				return false;
			}
			if (magnitude != prevqvalp.magnitude) {
				return false;
			}
		}
		return false;
	}

}
